# GNU Solfege - eartraining for GNOME
# Copyright (C) 2001  Tom Cato Amundsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import re
_token = re.compile(r"""
        (?P<newline>\n)
    |   (?P<space> \s+)
    |   (?P<comment>\#.*)
    |   (?P<name> [a-zA-Z_][\w-]*)
    |   (?P<number> \d+)
    |   (?P<punc> [-+=(){}/,\%])
    |   (?P<string> " [^"\\\n]* (?: \\. [^"\\\n]*)* "
        |           ' [^'\\\n]* (?: \\. [^'\\\n]*)* '
        )
""", re.VERBOSE).match
del re

(T_NEWLINE,
 T_SPACE,
 T_COMMENT,
 T_NAME,
 T_DONTCARE,
 T_NUMBER,
 T_PUNC,
 T_STRING,
 T_EOF,
)           = range(9)

# For debug output.
_enum2name = ["T_NEWLINE",
              "T_SPACE",
              "T_COMMENT",
              "T_NAME",
              "T_DONTCARE",
              "T_NUMBER",
              "T_PUNC",
              "T_STRING",
              "T_EOF",
             ]

_group2action = {
    "newline":  (T_NEWLINE, None),
    "space":    (T_SPACE, None),
    "comment":  (T_COMMENT, None),
    "name":     (T_NAME, None),
    "dontcare": (T_DONTCARE, None),
    "number":   (T_NUMBER, int),
    "punc":     (T_PUNC, None),
    "string":   (T_STRING, eval),
}
class ParseError:
    def __init__(self, a, b):
        self.m_a = a
        self.m_b = b
    def __str__(self):
        return "Error: '%s' '%s'" % (self.m_a, self.m_b)


import sys, os
if __name__ == "__main__":
    sys.path.append(".")
import gettext
import soundcard
from i18n import _
from soundcard.rat import Rat
import re, string
tokens = (
 'NAME', 'EQUALS', 'STRING', 'I18NSTRING', 'INTEGER', 'LBRACE',
 'RBRACE', 'COMMENT', 'PLUS', 'COMMA', 'PERCENT', 'DIVIDE',
 'BLOCKSTART', 'EQUALS_EXP', 'PLUS_EXP', 'DIVIDE_EXP', 'PERCENT_EXP',
 'COMMA_EXP', 'TEMPO', 'INCLUDE')

for t in tokens:
    globals()[t] = t


class DataparserException:
    def __init__(self, filename, lineno, badcode):
        self.filename = filename
        self.lineno = lineno
        self.badcode = badcode
    def __str__(self):
        return "dataparser-error, lineno: %i\nThis code is invalid: %s: %s" % (self.filename, self.lineno, self.badcode)


class Dataparser:
    def __init__(self, globals={}, gd=[]):
        self.m_globals = globals
        self.m_gd = gd
    def parse_file(self, filename):
        self.m_filename = [filename]
        v = []
        self.m_blocks = []
        context = self.m_globals
        f = open(filename, "r")
        s = f.read()
        f.close()
        import time
        t1 = time.clock()
        v = self.tokenize(s, v)
        print time.clock() - t1
        v = self.do_include(v)
        for op in (('%', PERCENT_EXP),
                   ('+', PLUS_EXP),
                   ('/', DIVIDE_EXP),
                   (',', COMMA_EXP),
                   ('=', EQUALS_EXP)):
            v = self.do_simple_op(v, op[0], op[1])
        v = self.do_blockstart(v)
        for ex in v:
            if ex[0] == EQUALS_EXP:
                context[ex[1][1]] = self.evaluate(ex[2])
            elif ex[0] == BLOCKSTART:
                assert context == self.m_globals
                self.m_blocks.append({})
                context = self.m_blocks[-1]
                context['blocktype'] = ex[1][1]
            elif ex[0] == T_PUNC and ex[1] == '}':
                for n in self.m_gd:
                    if not context.has_key(n):
                        context[n] = self.m_globals[n]
                context = self.m_globals
            elif ex[0] == T_STRING:
                # the shortcut that interprets a single string
                # "this is a string" as
                # music = "this is a string"
                context['music'] = ex[1]
            elif ex[0] == PERCENT_EXP:
                # the shortcut that interprets a single string
                # "this is a string" as
                # music = "this is a string"
                context['music'] = self.evaluate(ex)
            elif ex[0] == PLUS_EXP:
                # the shortcut that interprets a single string
                # "this is a string" as
                # music = "this is a string"
                context['music'] = self.evaluate(ex)
            else:
                raise "Parse error in '%s', line %i", ex
    def tokenize(self, s, list):
        i, n = 0, len(s)
        while i < n:
            m = _token(s, i)
            if not m:
                raise ParseError(s, i)
            group = m.lastgroup
            enum, action = _group2action[group]
            val = m.group(group)
            if action is not None:
                val = action(val)
            #self.token_eater(enum, val, list)
            if not (enum in (T_NEWLINE, T_SPACE, T_COMMENT)):
                list.append((enum, val, enum))
            i = m.end()
        #self.token_eater(T_EOF, None, list)
        return list
    def token_eater(self, t, val, list):
        if t == T_NEWLINE or t == T_SPACE or t == T_COMMENT:
            return
        list.append((t, val, _enum2name[t]))
        #list.append((_enum2name[t], val))
    def do_include(self, list):
        return list
        x = 0
        while x < len(list):
            if list[x][0] == INCLUDE:
                fn = os.path.join(os.path.dirname(self.m_filename[-1]),
                                  list[x][1])
                list[x:x+1] = self.get_tokens(fn)
            x = x + 1
        return list
    def do_simple_op(self, list, op, t):
        x = 1
        while x + 1 < len(list):
            if list[x][0] == T_PUNC and list[x][1] == op:
                list[x-1:x+2]=[(t, list[x-1], list[x+1])]
            else:
                x = x + 1
        return list
    def do_blockstart(self, list):
        x = 0
        while x < len(list):
            #FIXME muligens eget re pr tegn??
            if list[x][0] == T_NAME and list[x+1][0] == T_PUNC \
               and list[x+1][1] == '{':
                list[x:x+2]=[('BLOCKSTART', list[x])]
            else:
                x = x + 1
        return list
    def evaluate(self, exp):
        if exp[0] == DIVIDE_EXP:
            print "DIVIDE:", exp
            return Rat(int(exp[1][1]), int(exp[2][1]))
        elif exp[0] == T_STRING:
            return exp[1]
        elif exp[0] == I18NSTRING:
            return _(exp[1])
        elif exp[0] == T_NUMBER:
            return exp[1]
        elif exp[0] == PLUS_EXP:
            return self.evaluate(exp[1]) + self.evaluate(exp[2])
        elif exp[0] == PERCENT_EXP:
            if exp[1][0] is T_NAME:
                A = self.evaluate(exp[1])
            else:
                A = exp[1][1]
            return A % exp[2][1]
        elif exp[0] == COMMA_EXP:
            A = self.evaluate(exp[1])
            B = self.evaluate(exp[2])
            if type(A) == type([]):
                return A + [B]
            else:
                return [A, B]
            return [en]+[to]#[self.evaluate(exp[1]), self.evaluate(exp[2])]
        elif exp[0] == T_NAME:
            return self.m_globals[exp[1]]
        elif exp[0] == TEMPO:
            return exp[1]
        else:
            raise "unfinished", exp

def get_translated_string(dict, name):
    for n in gettext.lang:
        if dict.has_key("%s(%s)" % (name, n)):
            return dict["%s(%s)" % (name, n)]
    return dict[name]

def __f():
    if len(sys.argv) == 1:
        print "Give the file to parse as command line argument."
        sys.exit(-1)
    import time, os, os.path
    t1 = time.clock()
    for fn in sys.argv[1:]:
        if (not os.path.isfile(fn)) or (fn == 'lesson-files/Makefile'):
            continue
        p = Dataparser({'dictation': 'dictation',
                      'progression': 'progression',
                      'harmony': 'harmony',
                      'sing-chord': 'sing-chord',
                      'chord': 'chord',
                      'id-by-name': 'id-by-name',
                      'satb': 'satb',
                      'horiz': 'horiz',
                      'vertic': 'vertic',
                      'yes': 1,
                      'no': 0,
                      'tempo': (160, 4)})
        for x in range(10):
            p.parse_file(fn)
        #for b in p.m_blocks:
        #    print b, '\n'
    print time.clock() - t1


if __name__ == "__main__":
    import i18n
    i18n.initialise("share/locale")
    from i18n import _

    __f()
    sys.exit()
    import profile, pstats
    profile.run("__f()", "profile.txt")
    s = pstats.Stats("profile.txt")
    s.strip_dirs().sort_stats('cumulative').print_stats(100)
